import 'dart:async';
import 'dart:typed_data';
import 'dart:ui';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'dart:ui' as ui;
import 'package:flutter_cache_manager/flutter_cache_manager.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/color/nui_colors.dart';
import 'package:nui_core/data/nui_data_util.dart';

class NUIAvatarMarkerRenderer{
  NUIAvatarMarkerRenderer({this.color: NUIColors.NUISkeletalGray, this.shadowRadius: 10.0, this.textSpan, this.textBackground: NUIColors.NUIShadowBlackTransparent, this.shadowColor: NUIColors.NUIShadowBlackTransparent, this.outlineColor: NUIColors.NUIWhite, this.size, this.imagePath, this.thickness: 3, this.borderRadius});
  Color color;
  Color outlineColor;
  Color shadowColor;
  double shadowRadius;
  BorderRadius borderRadius;
  double thickness;
  double size;
  String imagePath;
  TextSpan textSpan;
  Color textBackground;

  Future<ui.Image> _loadImage(String photoUrl) async {
    final cache = DefaultCacheManager();
    final file = await cache.getSingleFile(photoUrl);
    final bytes = new Uint8List.fromList(await file.readAsBytes());

    final Completer<ui.Image> completer = new Completer();
    ui.decodeImageFromList(bytes, (ui.Image img) {
      logNUI("NUIMapMarkerRenderer", "Decoded image from path: $photoUrl, bytes : ${bytes.length}");
      return completer.complete(img);
    });
    return completer.future;
  }

  Future<Uint8List> drawAvatar(BuildContext context) async{
    final PictureRecorder pictureRecorder = PictureRecorder();
    final Canvas canvas = Canvas(pictureRecorder);
    final densityPixel = MediaQuery.of(context).devicePixelRatio;

    if(!isNullOrEmpty(imagePath)) {
      try {
        final image = await _loadImage(imagePath);

        NUIAvatarMarkerCustomPainter painter = NUIAvatarMarkerCustomPainter(
            color: color,
            outlineColor: outlineColor,
            thickness: thickness * densityPixel,
            image: image,
            shadowRadius: shadowRadius * densityPixel,
            shadowColor: shadowColor,
            textSpan: textSpan,
            textBackground: textBackground,
            densityRatio: densityPixel
        );

        painter.paint(canvas, Size(size * densityPixel, size * densityPixel));
      } catch (e) {
        logNUI("NUIMapMarkerRenderer", "Failed to load image with error : $e");
      }
    }
    else{
      NUIAvatarMarkerCustomPainter painter = NUIAvatarMarkerCustomPainter(
          color: color,
          outlineColor: outlineColor,
          thickness: thickness * densityPixel,
          shadowRadius: shadowRadius * densityPixel,
          shadowColor: shadowColor,
          textSpan: textSpan,
          textBackground: textBackground,
          densityRatio: densityPixel
      );

      painter.paint(canvas, Size(size * densityPixel, size * densityPixel));
    }

    final picture = pictureRecorder.endRecording();
    final pngBytes = await (await picture.toImage((size * densityPixel).toInt(), (size * densityPixel).toInt())).toByteData(format: ui.ImageByteFormat.png);

    return pngBytes.buffer.asUint8List();
  }

}

class NUIAvatarMarkerCustomPainter extends CustomPainter{
  NUIAvatarMarkerCustomPainter({this.color: NUIColors.NUISkeletalGray, this.shadowRadius: 10, this.textSpan, this.densityRatio: 1.0, this.textBackground: NUIColors.NUIShadowBlackTransparent, this.shadowColor: NUIColors.NUIShadowBlackTransparent, this.outlineColor: NUIColors.NUIWhite, this.image, this.thickness: 3, this.borderRadius});
  Color color;
  Color outlineColor;
  Color shadowColor;
  BorderRadius borderRadius;
  double thickness;
  double shadowRadius;
  ui.Image image;
  TextSpan textSpan;
  Color textBackground;
  double densityRatio;

  @override
  void paint(Canvas canvas, Size sizeBox) {
    final shadowRadius = (this.shadowRadius*2)/3*2;
    final size = sizeBox.width - (shadowRadius*2);
    final sizeWithShadow = sizeBox.width;
    final center = Offset(sizeWithShadow/2, sizeWithShadow/2);
    final radius = size/2;
    final trueShadowRadius = this.shadowRadius;

    // The circle should be paint before or it will be hidden by the path
    Paint paintCircle = Paint()
      ..color = color;

    Paint paintBorder = Paint()
      ..color = outlineColor
      ..strokeWidth = thickness
      ..style = PaintingStyle.stroke;

    Path shadowPath = Path()
      ..addOval(Rect.fromLTWH(shadowRadius, shadowRadius, size, size));
    canvas.drawShadow(shadowPath, shadowColor, trueShadowRadius/2, false);
    canvas.drawCircle(center, radius, paintCircle);
    canvas.drawCircle(center, radius - thickness/2, paintBorder);

    Path path = Path()
      ..addOval(Rect.fromLTWH(thickness + shadowRadius, thickness + shadowRadius, size - thickness - thickness, size - thickness - thickness));

    canvas.clipPath(path);
    if(image != null) {
      Size imgSize = Size(image.width.toDouble(), image.height.toDouble());
      Rect dstRect = Rect.fromLTWH(shadowRadius, shadowRadius, size, size);
      // Calculate the appropriate zoom size according to the adaptation mode
      FittedSizes fittedSizes = applyBoxFit(BoxFit.cover, imgSize, dstRect.size);
      // Get a Rect in the image area, at the specified size, at the centered position
      Rect inputRect = Alignment.center.inscribe(fittedSizes.source, Offset.zero & imgSize);
      // Get a Rect in the drawing area, at the specified size, at the centered position
      Rect outputRect = Alignment.center.inscribe(fittedSizes.destination, dstRect);
      canvas.drawImageRect(image, inputRect, outputRect, Paint());
    }

    final textSpan = this.textSpan;
    if(textSpan != null) {
      Paint textPaintCircle = Paint()
        ..color = textBackground;

      canvas.drawCircle(center, radius - thickness, textPaintCircle);

      final textPainter = TextPainter(text: textSpan, textDirection: TextDirection.ltr, textScaleFactor: densityRatio);
      textPainter.layout(minWidth: 0, maxWidth: sizeBox.width);
      textPainter.paint(canvas, Offset((sizeBox.width - textPainter.width) * 0.5, (sizeBox.height - textPainter.height) * 0.5,),);
    }
  }

  @override
  bool shouldRepaint(covariant CustomPainter oldDelegate) {
    return false;
  }

}