import 'dart:async';
import 'dart:typed_data';
import 'dart:ui';
import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:google_maps_cluster_manager/google_maps_cluster_manager.dart';
import 'package:google_maps_flutter/google_maps_flutter.dart';
import 'package:line_awesome_flutter/line_awesome_flutter.dart';
import 'package:nui_animation/nui_button_animation.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_geolocator/nui_geolocator.dart';
import 'package:nui_map/nui_map.dart';

typedef NUIMapViewMarkersCallback = void Function(List<NUIMapMarker> markers);
typedef NUIMapViewCameraPositionUpdate = void Function(LatLng target, double tilt, double zoom);
typedef NUIMapViewCirclesCallback = void Function(List<NUIMapCircle> circles);
typedef NUIMapViewIdCallback = void Function(String id);
typedef NUIMapViewRelocate = void Function({NUILatLng location});
typedef NUIMapViewLocationCallback = NUILatLng Function();
typedef NUIMapViewLocationUpdateListener = void Function(NUILatLng location);
typedef NUIMapViewOnMarkerTap = void Function(NUIMapMarker marker);
typedef NUIMapViewCameraPositionCallback = CameraPosition Function();

class NUIMapViewController{
  NUIMapViewController({this.destinationUpdateListener, this.markerTapListener});
  NUIMapViewMarkersCallback _addMarkersCallback;
  NUIMapViewMarkersCallback _addFixedMarkersCallback;
  NUIMapViewCameraPositionUpdate _cameraPositionUpdate;
  NUIMapViewCirclesCallback _addCirclesCallback;
  NUIMapViewIdCallback _removeMarkerCallback;
  NUIMapViewIdCallback _removeFixedMarkerCallback;
  NUIMapViewIdCallback _removeCircleCallback;
  NUIMapViewRelocate _relocateDestination;
  NUIMapViewRelocate _moveCameraCallback;
  Function _clearMarkersCallback;
  Function _clearFixedMarkersCallback;
  Function _clearCirclesCallback;
  NUIMapViewLocationCallback _latestLocationCallback;
  NUIMapViewLocationUpdateListener destinationUpdateListener;
  NUIMapViewOnMarkerTap markerTapListener;
  NUIMapViewCameraPositionCallback _cameraPositionCallback;

  void _onDestinationUpdate(NUILatLng location){
    if(destinationUpdateListener != null) destinationUpdateListener(location);
  }

  CameraPosition getCurrentCameraPosition(){
    if(_cameraPositionCallback != null) return _cameraPositionCallback();
    return null;
  }

  void setCameraPosition(LatLng target, double tilt, double zoom){
    if(_cameraPositionUpdate != null) _cameraPositionUpdate(target, tilt, zoom);
  }

  void _onMarkerTap(NUIMapMarker marker){
    if(markerTapListener != null) markerTapListener(marker);
  }

  NUILatLng getLatestDestinationLocation(){
    if(_latestLocationCallback != null) return _latestLocationCallback();
    return null;
  }

  void relocateDestination({NUILatLng location}){
    if(_relocateDestination != null) _relocateDestination(location: location);
  }
  void moveCamera(NUILatLng location){
    if(_moveCameraCallback != null) _moveCameraCallback(location: location);
  }
  void clearMarkers(){
    if(_clearMarkersCallback != null) _clearMarkersCallback();
  }
  void clearFixedMarkers(){
    if(_clearFixedMarkersCallback != null) _clearFixedMarkersCallback();
  }
  void clearCircles(){
    if(_clearCirclesCallback != null) _clearCirclesCallback();
  }
  void removeMarker(String id){
    if(_removeMarkerCallback != null) _removeMarkerCallback(id);
  }
  void removeFixedMarker(String id){
    if(_removeFixedMarkerCallback != null) _removeFixedMarkerCallback(id);
  }
  void removeCircle(String id){
    if(_removeCircleCallback != null) _removeCircleCallback(id);
  }
  void addMarkers(List<NUIMapMarker> markers){
    if(_addMarkersCallback != null) _addMarkersCallback(markers);
  }
  void addFixedMarkers(List<NUIMapMarker> markers){
    if(_addFixedMarkersCallback != null) _addFixedMarkersCallback(markers);
  }
  void addCircles(List<NUIMapCircle> circles){
    if(_addCirclesCallback != null) _addCirclesCallback(circles);
  }
}

class NUIMapView extends StatefulWidget {
  NUIMapView({this.mapType: MapType.normal, this.enableZoom: true, this.initialZoom: 16.0, this.gestureRecognizers, this.tilt : 0.0, this.relocateButtonSize: 32, this.padding, this.currentLocation, this.enableRelocate: false, this.liteMode: false, this.destination, this.enableCompass: false, this.fixedMarkers, this.markers, this.circles, this.controller});
  final MapType mapType;
  final bool enableZoom;
  final bool enableCompass;
  final EdgeInsets padding;
  final bool liteMode;
  final double initialZoom;
  final double tilt;
  final NUIMapMarker destination;
  final NUIMapMarker currentLocation;
  final List<NUIMapMarker> markers;
  final List<NUIMapMarker> fixedMarkers;
  final List<NUIMapCircle> circles;
  final NUIMapViewController controller;
  final bool enableRelocate;
  final double relocateButtonSize;
  final Set<Factory<OneSequenceGestureRecognizer>> gestureRecognizers;

  @override
  State<NUIMapView> createState() => NUIMapViewState();
}

class NUIMapViewState extends State<NUIMapView> {
  final _componentId = randomUUID();

  Uint8List customMarker;
  Completer<GoogleMapController> _controller = Completer();
  final LatLng _center = const LatLng(45.521563, -122.677433);

  NUILatLng _currentPosition;
  NUIMapViewController _mapController;
  ClusterManager _manager;

  NUIMapMarker _destinationMarker;
  NUIMapMarker _currentLocationMarker;

  Map<String, NUIMapMarker> _markerList = Map();
  Map<String, NUIMapMarker> _fixedMarkerList = Map();
  Map<String, NUIMapCircle> _circleList = Map();

  final Map<String, Marker> _markers = Map();
  final Map<String, Marker> _fixedMarkers = Map();
  final Map<String, Circle> _circles = Map();

  CameraPosition _currentCameraPosition;

  @override
  void initState(){
    super.initState();
    _destinationMarker = widget.destination;
    _currentLocationMarker = widget.currentLocation;

    widget.markers?.forEach((element) {
      _markerList[element.id] = element;
    });
    widget.fixedMarkers?.forEach((element) {
      _fixedMarkerList[element.id] = element;
    });
    widget.circles?.forEach((element) {
      _circleList[element.id] = element;
    });

    _mapController = widget.controller ?? NUIMapViewController();

    _mapController._addCirclesCallback = (circles){
      _addCircles(circles);
    };
    _mapController._addMarkersCallback = (markers){
      _addMarkers(markers);
    };
    _mapController._addFixedMarkersCallback = (markers){
      _addFixedMarkers(markers);
    };
    _mapController._clearMarkersCallback = (){
      _clearMarkers();
    };
    _mapController._clearFixedMarkersCallback = (){
      _clearFixedMarkers();
    };
    _mapController._clearCirclesCallback = (){
      _clearCircles();
    };
    _mapController._removeMarkerCallback = (id){
      _removeMarker(id);
    };
    _mapController._removeFixedMarkerCallback = (id){
      _removeFixedMarker(id);
    };
    _mapController._removeCircleCallback = (id){
      _removeCircle(id);
    };
    _mapController._relocateDestination = ({location}){
      _moveDestinationToCurrentLocation(newPosition: location);
    };
    _mapController._moveCameraCallback = ({location}){
      _moveToPosition(position: CameraPosition(target: LatLng(location.latitude, location.longitude), tilt: _currentCameraPosition.tilt, zoom: _currentCameraPosition.zoom));
    };
    _mapController._cameraPositionUpdate = (target, tilt, zoom){
      _moveToPosition(position: CameraPosition(target: target, tilt: tilt, zoom: zoom));
    };
    _mapController._cameraPositionCallback = () => _currentCameraPosition;
    _mapController._latestLocationCallback = (){
      return _destinationMarker?.position;
    };

    _manager = _initClusterManager(widget.markers);
  }

  List<NUIMapClusterItem> _getClusterItems(List<NUIMapMarker> markers){
    logNUI("NUIMapView", "Getting cluster items of size : ${markers.length}");
    return markers.map((e) {
      return NUIMapClusterItem(LatLng(e.position.latitude, e.position.longitude), e);
    }).toList();
  }

  ClusterManager _initClusterManager(List<NUIMapMarker> markers) {
    return ClusterManager<NUIMapMarker>(
      _getClusterItems(_getAllClusterMarkers()),
      _updateClusterMarkers,
      markerBuilder: _markerBuilder,
      initialZoom: widget.initialZoom ?? 16
    );
  }

  //Handled by cluster manager
  void _updateClusterMarkers(Set<Marker> markers){
    logNUI("NUIMapView", "Updating cluster markers of size : ${markers.length}");
    _markers.clear();
    for(Marker marker in markers) {
      logNUI("NUIMapView", "Marker is null : ${marker == null}");
      _markers[marker.markerId.value] = marker;
    }
    setState(() {

    });
  }

  Future<Marker> Function(Cluster<NUIMapMarker>) get _markerBuilder => (cluster) async {
    final clusterCount = cluster.count;
    if(clusterCount > 1){
      logNUI("NUIMapView", "Found cluster of size $clusterCount");
      final clusterItems = <NUIMapMarker>[];
      clusterItems.addAll(cluster.items);

      final marker = cluster.items.first;
      return marker.getClusteredMarker(clusterItems, onTap: (){
        logNUI("NUIMapView", "Clicked on cluster");
        _onMarkerTap(marker, isCluster: true, clusterItems: clusterItems);
      }) ?? marker.getMarker(onTap: (){
        _onMarkerTap(marker);
      });
    }
    else{
      final marker = cluster.items.first;
      return marker.getMarker(onTap: (){
        _onMarkerTap(marker);
      });
    }
  };

  List<NUIMapMarker> _getAllClusterMarkers(){
    List<NUIMapMarker> markers = [];

    _markerList.forEach((key, value) {
      markers.add(value);
    });

    logNUI("NUIMapView", "Total of ${markers.length} markers found to render before cluster");
    return markers;
  }

  void _refreshClusterItems(){
    _manager.setItems(_getClusterItems(_getAllClusterMarkers()));
  }

  void _refreshCircles(){
    _refreshRenderCircleList();
    setState(() {

    });
  }

  void _refreshFixedMarkers() async{
    await _refreshRenderFixedMarkerList();
    setState(() {
      
    });
  }

  //Methods for altering the circles and markers
  void _addMarkers(List<NUIMapMarker> markers){
    markers.forEach((element) {
      _markerList[element.id] = element;
    });
    _refreshClusterItems();
  }

  void _addFixedMarkers(List<NUIMapMarker> markers){
    markers.forEach((element) {
      _fixedMarkerList[element.id] = element;
    });
    _refreshFixedMarkers();
  }

  void _addCircles(List<NUIMapCircle> circles){
   circles.forEach((element) {
     _circleList[element.id] = element;
   });
   _refreshCircles();
  }

  void _clearMarkers(){
    _markerList.clear();
    _refreshClusterItems();
  }

  void _clearFixedMarkers(){
    _fixedMarkerList.clear();
    _refreshFixedMarkers();
  }

  void _clearCircles(){
    _circleList.clear();
    _refreshCircles();
  }

  void _removeMarker(String id){
    _markerList.remove(id);
    _refreshClusterItems();
  }

  void _removeFixedMarker(String id){
    _fixedMarkerList.remove(id);
    _refreshFixedMarkers();
  }

  void _removeCircle(String id){
    _circleList.remove(id);
    _refreshCircles();
  }

  Future<bool> _refreshRenderFixedMarkerList() async{
    _fixedMarkers.clear();

    for(String key in _fixedMarkerList.keys){
      final value = _fixedMarkerList[key];
      _fixedMarkers[key] = await value.getMarker(onTap: (){
        _onMarkerTap(value);
      });
    }
    if(_destinationMarker != null){
      _fixedMarkers[_destinationMarker.id] = await _destinationMarker.getMarker(onTap: (){
        _onMarkerTap(_destinationMarker);
      });
    }
    if(_currentLocationMarker != null){
      _fixedMarkers[_currentLocationMarker.id] = await _currentLocationMarker.getMarker(onTap: (){
        _onMarkerTap(_currentLocationMarker);
      });
    }
    return true;
  }

  Future<bool> _refreshRenderCircleList() async {
    _circles.clear();

    _circleList.forEach((key, value) {
      _circles[key] = value.getCircle();
    });
    return true;
  }

  CameraPosition _getCameraPosition(LatLng latLng, {double zoom}){
    return CameraPosition(target: latLng, zoom: zoom ?? widget.initialZoom ?? 16, tilt: widget.tilt ?? 0);
  }

  void _moveDestinationToCurrentLocation({NUILatLng newPosition}) async{
    if(_destinationMarker == null && _currentLocationMarker == null) return;
    final curLocation = newPosition ?? await NUIGeoLocator.get().getLastKnownLatLng();
    if(curLocation != null){
      _destinationMarker?.position = curLocation;
      _currentLocationMarker?.position = curLocation;
      if(_destinationMarker != null) {
        _mapController._onDestinationUpdate(curLocation);
      }
      _moveToPosition(position: _getCameraPosition(LatLng(curLocation.latitude, curLocation.longitude)));
      _refreshClusterItems();
    }
  }

  void _focusDestination() async{
    if(_destinationMarker == null) return;
    _moveToPosition(position: _getCameraPosition(LatLng(_destinationMarker.position.latitude, _destinationMarker.position.longitude)));
  }

  void _focusCurrentLocation() async{
    final curLocation = await NUIGeoLocator.get().getLastKnownLatLng();
    if(curLocation != null){
      logNUI("NUIMapView", "Drawing current location marker");
      _currentPosition = curLocation;
      _currentLocationMarker..position = curLocation;
      _moveToPosition(position: _getCameraPosition(LatLng(_currentPosition.latitude, _currentPosition.longitude)));
    }
  }

  void _onMarkerTap(NUIMapMarker marker, {bool isCluster: false, List<NUIMapMarker> clusterItems}){
    logNUI("NUIMapView", "On marker tapped, data : ${marker.data}");
    _mapController._onMarkerTap(marker);
    if(isCluster == true){
      //Tapped on cluster, expand the cluster?
      logNUI("NUIMapView", "On cluster clicked");
      _expandCluster(clusterItems);
    }
  }

  LatLng _calculateCenterOfCluster(List<NUIMapMarker> clusterItems){
    double totalLat = 0;
    double totalLng = 0;
    clusterItems.forEach((element) {
      totalLat += element.position.latitude;
      totalLng += element.position.longitude;
    });

    double centerLat = totalLat / clusterItems.length;
    double centerLng = totalLng / clusterItems.length;
    return LatLng(centerLat, centerLng);
  }

  void _expandCluster(List<NUIMapMarker> clusterItems) async{
    if(_currentCameraPosition == null) return;

    final centerLatLng = _calculateCenterOfCluster(clusterItems);
    final curZoom = _currentCameraPosition.zoom;
    final nextZoom = curZoom + 3;

    _moveToPosition(position: _getCameraPosition(centerLatLng, zoom: nextZoom));
  }

  void _onMapCreated(GoogleMapController controller) async{
    if(_destinationMarker != null){
      _focusDestination();
    }
    else if(_currentLocationMarker != null) {
      _focusCurrentLocation();
    }

    logNUI("NUIMapView", "Google map created");
    await _refreshRenderFixedMarkerList();
    await _refreshRenderCircleList();
    _controller.complete(controller);
    _manager.setMapController(controller);
  }

  @override
  Widget build(BuildContext context) {
    final markers = <Marker>[];
    markers.addAll(_markers.values);
    markers.addAll(_fixedMarkers.values);

    final circles = <Circle>[];
    circles.addAll(_circles.values);

    return Material(
      child: Stack(
        children: [
          GoogleMap(
            mapType: widget.mapType ?? MapType.normal,
            initialCameraPosition: _currentPosition != null ? _getCameraPosition(LatLng(_currentPosition.latitude, _currentPosition.longitude)) : CameraPosition(target: _center),
            zoomControlsEnabled: false,
            markers: markers.toSet(),
            circles: circles.toSet(),
            liteModeEnabled: widget.liteMode == true,
            zoomGesturesEnabled: widget.enableZoom ?? false,
            compassEnabled: widget.enableCompass ?? false,
            mapToolbarEnabled: false,
            padding: widget.padding ?? EdgeInsets.all(0),
            onMapCreated: (GoogleMapController controller) {
              _onMapCreated(controller);
            },
            gestureRecognizers: widget.gestureRecognizers,
            onCameraIdle: (){
              _manager.updateMap();
            },
            onCameraMove: (position){
              _currentCameraPosition = position;
              _manager.onCameraMove(position);
            },
            onCameraMoveStarted: (){

            },
          ),
          Positioned(
            bottom: 5,
            right: 5,
            child: Visibility(
              visible: widget.enableRelocate == true,
              child: Padding(
                padding: widget.padding ?? EdgeInsets.all(0),
                child: NUIBouncingAnimation(
                    child: Container(
                        height: 36,
                        width: 36,
                        alignment: Alignment.center,
                        decoration: NUIBoxDecoration(
                            color: NUIColors.NUIWhite.withOpacity(0.5),
                            borderRadius: BorderRadius.circular(18)
                        ),
                        child:  Container(
                          height: widget.relocateButtonSize ?? 32,
                          width: widget.relocateButtonSize ?? 32,
                          alignment: Alignment.center,
                          decoration: NUIBoxDecoration(
                              gradient: NUILinearGradient(
                                colors: [
                                  NUIColors.NUIWhite.withOpacity(0.5),
                                  NUIColors.NUIWhite.withOpacity(0.7)
                                ],
                                orientation: GradientOrientation.TL_BR
                              ),
                              borderRadius: BorderRadius.circular((widget.relocateButtonSize ?? 32)/2)
                          ),
                          child: Icon(LineAwesomeIcons.location_arrow, size: (widget.relocateButtonSize ?? 32) - 10, color: NUIColors.NUITextBlack,)
                        )
                    ),
                    onTap: (){
                      _moveDestinationToCurrentLocation();
                    }
                ),
              ),
            ),
          )
        ],
      ),
    );
  }

  Future<void> _moveToPosition({bool animate: true, CameraPosition position}) async {
    final GoogleMapController controller = await _controller.future;
    if(animate == true){
      logNUI("NUIMapView", "Animating to position : $position");
      controller.animateCamera(CameraUpdate.newCameraPosition(position));
    }
    else{
      logNUI("NUIMapView", "Moving to position : $position");
      controller.moveCamera(CameraUpdate.newCameraPosition(position));
    }
  }
}
